#include <Arduino.h>
#include <pico.h>
#include <hardware/pwm.h>
#include <Keyboard.h>

#define PORT 2
#define PORT2 3
#define THRESHOLD 10

class HallButton
{
public:
  uint64_t firstPulseStart;
  uint8_t counter = 0;
  uint64_t firstPulseLength[3]; // value of button
  int64_t value;
  uint64_t _median;
  bool isDataReadyToRead = false;
  bool isButtonPressed = false;
  uint64_t nextTogglePoint = 0; // value when to toggle state of button

  HallButton(uint8_t gpioNum, char buttonOnKeyboard)
  {
    this->gpioNum = gpioNum;
    this->buttonOnKeyboard = buttonOnKeyboard;
  }

  void HandleISR(bool gpioState, uint64_t currentTime)
  {
    if (gpioState)
    {
      firstPulseLength[counter] = currentTime - firstPulseStart;
      isDataReadyToRead = true;
      if (++counter >= 3)
        counter = 0;
    }
    else
    {
      firstPulseStart = currentTime;
    }
  }
  void HandleButton()
  {
    if (!isDataReadyToRead)
      return;
    _median = median(firstPulseLength[0], firstPulseLength[1], firstPulseLength[2]);
    value = filtered((float)_median);
    if (nextTogglePoint == 0)
      nextTogglePoint = value + THRESHOLD;
    if (isButtonPressed)
    {
      if (value < nextTogglePoint)
      {
        Release();
      }
      if ((int64_t)value - (int64_t)nextTogglePoint > THRESHOLD)
        nextTogglePoint = value - THRESHOLD;
    }
    else
    {
      if (value >= nextTogglePoint)
      {
        Press();
      }
      if ((int64_t)nextTogglePoint - (int64_t)value > THRESHOLD)
        nextTogglePoint = value + THRESHOLD;
    }
    isDataReadyToRead = false;
  }
  uint64_t median(uint64_t a, uint64_t b, uint64_t c)
  {
    return (a < b) ? ((b < c) ? b : ((c < a) ? a : c)) : ((a < c) ? a : ((c < b) ? b : c));
  }
  float filtered(float value)
  {
    float _kalman_gain, _current_estimate;
    _kalman_gain = _err_estimate / (_err_estimate + _err_measure);
    _current_estimate = _last_estimate + _kalman_gain * (value - _last_estimate);
    _err_estimate = (1.0 - _kalman_gain) * _err_estimate + fabs(_last_estimate - _current_estimate) * _q;
    _last_estimate = _current_estimate;
    return _current_estimate;
  }

private:
  float _err_measure = 15;
  float _err_estimate = 15;
  float _q = 0.1;
  float _last_estimate = 0.0;
  uint8_t gpioNum;
  char buttonOnKeyboard;
  void Press()
  {
    if (isButtonPressed)
      return;
    Keyboard.press(buttonOnKeyboard);
    isButtonPressed = true;
  }
  void Release()
  {
    if (!isButtonPressed)
      return;
    Keyboard.release(buttonOnKeyboard);
    isButtonPressed = false;
  }
};

HallButton btn[2] = {HallButton(PORT, 't'), HallButton(PORT2, 'y')};

void isr(uint gpio, uint32_t events)
{
  uint64_t currentTime = time_us_64();
  bool state = events == 8;
  switch (gpio)
  {
  case PORT:
    btn[0].HandleISR(state, currentTime);
    break;
  case PORT2:
    btn[1].HandleISR(state, currentTime);
    break;
  }
}

void setup()
{
  gpio_init(PORT);
  gpio_set_dir(PORT, false);
  gpio_pull_up(PORT);
  gpio_set_irq_enabled_with_callback(PORT, GPIO_IRQ_EDGE_RISE | GPIO_IRQ_EDGE_FALL, true, isr);
  gpio_init(PORT2);
  gpio_set_dir(PORT2, false);
  gpio_pull_up(PORT2);
  gpio_set_irq_enabled_with_callback(PORT2, GPIO_IRQ_EDGE_RISE | GPIO_IRQ_EDGE_FALL, true, isr);
  // Serial.begin(9600);
  Keyboard.begin();
}

void printDebugInfo()
{
  Serial.printf("%llu,%llu,%llu,%llu,%i,", btn[0].firstPulseLength[btn[0].counter], btn[0]._median, btn[0].value, btn[0].nextTogglePoint, btn[0].isButtonPressed);
  Serial.printf("%llu,%llu,%llu,%llu,%i\n", btn[1].firstPulseLength[btn[1].counter], btn[1]._median, btn[1].value, btn[1].nextTogglePoint, btn[1].isButtonPressed);
}
void loop()
{
  btn[0].HandleButton();
  btn[1].HandleButton();
}